/*=============================================================================
vitalnix/libadb - User Management Suite - Unified Account Database
groupadd.c - Create a new group
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
  E_SUCCESS = 0,
  E_OTHER,       // other error, see errno
  E_OPEN,        // unable to open UDB module or DB
  E_GID_USED,    // GID already used and -o was not specified
  E_NAME_USED,   // group already exists
  E_RANGE,       // GID is not in the normal-range and -r was not specified
  E_UPDATE,      // db->groupadd() did not return ok
  E_CLOSE,       // db->close() did not return ok
};

struct opt {
  long gid_min, gid_max;
} conf;

struct {
  char *gname;
  long gid;

  // groupadd internal
  int dup, inter, sys;
} opt;

static int count_args(const char **);
static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  struct accdb_module *db;
  struct accdb_group mmask, smask;
  int eax, rv = E_SUCCESS;

  read_config();

  memset(&opt, 0, sizeof(opt));
  memset(&mmask, 0, sizeof(struct accdb_group));
  memset(&smask, 0, sizeof(struct accdb_group));
  opt.gid = -1;

  if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

  if((db = accdb_load(module_path)) == NULL) {
    fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
     module_path, strerror(errno));
    return E_OPEN;
  }

  opt.gname = *argv;
  if(opt.inter) {
    const void *qtab[] = {
      "Group name\n[%s] > ", &opt.gname,
      "Group ID (-1 for next free)\n[%d] > ", &opt.gid,
      NULL,
    };

    p_interactive(qtab);
  }

  if((eax = db->open(&db->state)) <= 0) {
    fprintf(stderr, "Could not open ACCDB back-end: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OPEN;
    goto __main__close_accdb;
  }

  smask.gname = mmask.gname = opt.gname;
  smask.gid = -1;

  if((eax = db->groupinfo(db->state, &smask, NULL)) < 0) {
    fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OTHER;
    goto __main__close_backend;
  }

  if(eax > 0) {
    fprintf(stderr, "Group \"%s\" already exists\n", smask.gname);
    rv = E_NAME_USED;
    goto __main__close_backend;
  }

  if(opt.gid != -1) { // -g is provided
    // The -o flag (allow creating group with duplicate GID) was not passed
    smask.gname = NULL;
    smask.gid = opt.gid;
    if(opt.gid >= 0 && !opt.dup &&
     db->groupinfo(db->state, &smask, NULL) > 0) {
      fprintf(stderr, "Group with GID %ld already exists."
       " Use -o to override.\n", smask.gid);
      rv = E_GID_USED;
      goto __main__close_backend;
    }
    mmask.gid = opt.gid;
  } else if(opt.sys) { // -r flag passed
    mmask.gid = db->sysctl(ACCDB_NEXTGID_SYS, db->state);
  } else {
    mmask.gid = db->sysctl(ACCDB_NEXTGID, db->state);
  }

  if((eax = db->groupadd(db->state, &mmask)) <= 0) {
    fprintf(stderr, "Warning: Group addition failed: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_UPDATE;
  }

__main__close_backend:
  if((eax = db->close(db->state)) <= 0) {
    fprintf(stderr, "Warning: DB was not cleanly closed: %s"
     " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
    if(rv == E_SUCCESS) { rv = E_CLOSE; }
  }

__main__close_accdb:
  accdb_unload(db);
  return rv;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
  int argk = 0;
  while(*args++ != NULL) { ++argk; }
  return argk;
}

static int get_options(int *argc, const char ***argv) {
  static const char *_empty_argv[] = {NULL};
  const struct poptOption options_table[] = {
    // New, uudb-groupadd options
    {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
     "Interactively prompt for parameters", NULL},
    {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
     "Use a different module than \"*\" (the default)", "name"},

    // Default options
    {NULL, 'g', POPT_ARG_INT, &opt.gid, '\0',
     "Numerical value of the group's ID (UDB module might ignore this)",
     "gid"},
    {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
     "Allow creating a group with non-unique GID"
     " (might be disabled by UDB module)", NULL},
    {NULL, 'r', POPT_ARG_NONE, &opt.sys, '\0',
     "System group (allow/use GID < GID_MIN or GID > GID_MAX)", NULL},

    POPT_AUTOHELP
    POPT_TABLEEND
  };

  poptContext ctx;
  const char **args;
  int c, argk;

  ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
  poptSetOtherOptionHelp(ctx, "[OPTIONS...] group");
  while((c = poptGetNextOpt(ctx)) >= 0);
  if(c < -1) {
    fprintf(stderr, "%s: %s: %s\n", **argv, poptBadOption(ctx, 0),
     poptStrerror(c));
    poptFreeContext(ctx);
    return 0;
  }

  if((args = poptGetArgs(ctx)) != NULL) {
    argk = count_args(args);
    poptDupArgv(argk, args, argc, argv);
  } else {
    *argc = 0;
    *argv = _empty_argv;
  }

  if(!opt.inter && **argv == NULL) {
    // gname is mandatory
    poptPrintUsage(ctx, stderr, 0);
    poptFreeContext(ctx);
    return 0;
  }

  poptFreeContext(ctx);
  return 1;
}

static void read_config(void) {
  const char *paths[] = {"etc", "/etc", NULL};
  struct rconfig_opt table[] = {
    {"GID_MIN", 'l', &conf.gid_min},
    {"GID_MAX", 'l', &conf.gid_max},
    {NULL, '\0', NULL},
  };

  // First, our hardcoded values
  conf.gid_min = 100;
  conf.gid_max = 65000;

  // Then read from files
  accdb_rconfig_pv(paths, "autouid", table, RCONFIG_ONE);
  return;
}

//==[ End of file ]============================================================
